"""
### Pushed Over
> Typed, async client for the Pushover API
"""
from ._types import HighPriorityParams, LowPriorityParams, Ok, PriorityOk, PushoverError, Err, ValidationError
from .main import notify


from typing_extensions import TypedDict, Literal, NotRequired, TypeAlias
from dataclasses import dataclass
from pydantic import BaseModel, ValidationError as PyValidationError
import httpx

class Params(TypedDict):
  message: str
  title: NotRequired[str]
  device: NotRequired[str]
  url: NotRequired[str]
  url_title: NotRequired[str]
  ttl: NotRequired[int]
  """Seconds after which the message will be deleted"""
  sound: NotRequired[str]

class HighPriorityParams(Params):
  priority: Literal[2]
  expire: int
  """Quoting the [docs](https://pushover.net/api): how many seconds your notification will continue to be retried for (every retry seconds). If the notification has not been acknowledged in expire seconds, it will be marked as expired and will stop being sent to the user. Note that the notification is still shown to the user after it is expired, but it will not prompt the user for acknowledgement. This parameter must have a maximum value of at most 10800 seconds (3 hours), though the total number of retries will be capped at 50 regardless of the expire parameter."""
  retry: int
  """Quoting the [docs](https://pushover.net/api): how often (in seconds) the Pushover servers will send the same notification to the user. In a situation where your user might be in a noisy environment or sleeping, retrying the notification (with sound and vibration) will help get his or her attention. This parameter must have a value of at least 30 seconds between retries."""

class LowPriorityParams(Params):
  priority: NotRequired[Literal[-2, -1, 0, 1]]

class Ok(BaseModel):
  status: Literal[1]
  request: str

class PriorityOk(Ok):
  receipt: str

class PushoverError(BaseModel):
  status: int
  errors: list[str]
  request: str
  tag: Literal['pushover'] = 'pushover'

@dataclass
class NetworkError:
  err: httpx.HTTPError
  tag: Literal['network'] = 'network'

@dataclass
class ValidationError:
  err: PyValidationError
  tag: Literal['validation'] = 'validation'

Err = PushoverError | NetworkError | ValidationError


from typing_extensions import Unpack, overload
import os
import httpx
from haskellian import either as E, Left, Either
from ._types import HighPriorityParams, LowPriorityParams, Ok, PriorityOk, PushoverError, Err

@overload
async def notify(*, token: str | None = None, user: str | None = None, attachment: bytes | None = None, **params: Unpack[HighPriorityParams]) -> Either[Err, PriorityOk]:
  """Send a high-priority notification
  - `token`: can be `None` if the `PUSHOVER_TOKEN` env. variable is set
  - `user`: can be `None` if the `PUSHOVER_USER` env. variable is set
  - `attachment`: an image to send with the notification
  
  This function doesn't raise exceptions (except if you don't provide a token), but instead returns an appropriate `Left` value
  """
@overload
async def notify(*, token: str | None = None, user: str | None = None, attachment: bytes | None = None, **params: Unpack[LowPriorityParams]) -> Either[Err, Ok]:
  """Send a non-urgent notification
  - `token`: can be `None` if the `PUSHOVER_TOKEN` env. variable is set
  - `user`: can be `None` if the `PUSHOVER_USER` env. variable is set
  - `attachment`: an image to send with the notification
  
  This function doesn't raise exceptions (except if you don't provide a token), but instead returns an appropriate `Left` value
  """

async def notify(*, token: str | None = None, user: str | None = None, attachment: bytes | None = None, **params): # type: ignore
  token = token or os.getenv('PUSHOVER_TOKEN')
  user = user or os.getenv('PUSHOVER_USER')
  if not token:
    raise ValueError('Please provide a token or set the PUSHOVER_TOKEN env. variable')
  if not user:
    raise ValueError('Please provide a user or set the PUSHOVER_USER env. variable')
  
  params = { **params, 'token': token, 'user': user }
  files = {'attachment': attachment} if attachment else None
  
  try:
    async with httpx.AsyncClient() as client:
      response = await client.post('https://api.pushover.net/1/messages.json', data=params, files=files)
      if response.status_code == 200:
        model = PriorityOk if params.get('priority') == 2 else Ok
        return E.validate_json(response.text, model)
      else:
        return E.validate_json(response.text, PushoverError).bind(Left)
  except httpx.HTTPError as e:
    return Left(e)